import uuid
from unittest import TestCase

from musicscore import Part
from musicscore.beat import (
    Beat,
    _convert_to_quarter_duration_splittables_dictionary,
    get_chord_group_subdivision,
)
from musicscore.chord import Chord
from musicscore.config import SPLITTABLES
from musicscore.exceptions import (
    AddChordError,
    VoiceIsFullError,
    BeatNotFullError,
    QuarterDurationIsNotWritable,
)
from musicscore.measure import Measure
from musicscore.quarterduration import QuarterDuration
from musicscore.staff import Staff

# from unittest import TestCase
from musicscore.tuplet import Tuplet
from musicscore.voice import Voice


def create_voice():
    voice = Voice()
    st = Staff()
    m = Measure(1)
    m.add_child(st)
    st.add_child(voice)
    return voice


class TestBeat(TestCase):
    def test_beat_quarter_duration(self):
        b = Beat()
        assert b.quarter_duration == 1
        b.quarter_duration = 4
        assert b.quarter_duration == 4

    def test_beat_previous(self):
        b = Beat()
        assert b.previous is None
        v = create_voice()
        v.update_beats(1 / 4, 1 / 8, 1 / 2, 1)
        assert v.get_children()[0].previous is None
        assert v.get_children()[1].previous.quarter_duration == 1 / 4
        assert v.get_children()[2].previous.quarter_duration == 1 / 8
        assert v.get_children()[3].previous.quarter_duration == 1 / 2

    def test_beat_next(self):
        b = Beat()
        assert b.next is None
        v = create_voice()

        v.update_beats(1 / 4, 1 / 8, 1 / 2, 1)
        assert v.get_children()[0].next.quarter_duration == 1 / 8
        assert v.get_children()[1].next.quarter_duration == 1 / 2
        assert v.get_children()[2].next.quarter_duration == 1
        assert v.get_children()[3].next is None

    def test_beat_offset(self):
        b = Beat()
        assert b.offset is None
        v = create_voice()
        v.update_beats(1 / 4, 1 / 8, 1 / 2, 1)
        assert v.get_children()[0].offset == 0
        assert v.get_children()[1].offset == 1 / 4
        assert v.get_children()[2].offset == 1 / 4 + 1 / 8
        assert v.get_children()[3].offset == 1 / 4 + 1 / 8 + 1 / 2

    def test_beat_fill_with_rest(self):
        b = Beat()
        b._parent = Voice()
        assert not b.is_filled
        b.fill_with_rests()
        assert b.is_filled

    def test_get_chord_groups_subdivision(self):
        assert (
            get_chord_group_subdivision([Chord(60, qd) for qd in [2 / 5, 2 / 5, 1 / 5]])
            == 5
        )
        assert (
            get_chord_group_subdivision([Chord(60, qd) for qd in [4 / 5, 4 / 5, 2 / 5]])
            == 5
        )
        assert (
            get_chord_group_subdivision(
                [Chord(60, qd) for qd in [1 / 5, 1 / 5, 1 / 10]]
            )
            == 5
        )
        assert (
            get_chord_group_subdivision(
                [Chord(60, qd) for qd in [1 / 6, 1 / 6, 1 / 6, 1 / 10, 3 / 10, 1 / 10]]
            )
            is None
        )


class TestBeatUpdateChords(TestCase):
    def test_beat_update_chord_types(self):
        p = Part("p1")
        [p.add_chord(ch) for ch in [Chord(60, 0.5), Chord(70, 0.5)]]
        beat = p.get_beat(
            measure_number=1, staff_number=1, voice_number=1, beat_number=1
        )
        assert {ch.type for ch in beat.get_chords()} == {None}
        beat._update_chord_types()
        assert {ch.type for ch in beat.get_chords()} == {"eighth"}
        [p.add_chord(ch) for ch in [Chord(60, 0.5), Chord(70, 0.5)]]
        beat = p.get_beat(
            measure_number=1, staff_number=1, voice_number=1, beat_number=2
        )
        beat.get_chords()[0].type = "quarter"
        beat._update_chord_types()
        assert [ch.type for ch in beat.get_chords()] == ["quarter", "eighth"]

    def test_beat_update_chord_number_of_dots(self):
        p = Part("p1")
        [p.add_chord(Chord([60, 61], qd)) for qd in [1, 0.75, 0.25, 0.875, 0.125]]
        assert {ch.number_of_dots for ch in p.get_chords()} == {None}
        for b in p.get_beats():
            try:
                b._update_chord_number_of_dots()
            except BeatNotFullError:
                pass
        assert [ch.number_of_dots for ch in p.get_chords()] == [0, 1, 0, 2, 0]
        p = Part("p2")
        ch = Chord(60, 1)
        ch.number_of_dots = 3
        p.add_chord(ch)
        beat = p.get_beats()[0]
        beat._update_chord_number_of_dots()
        assert beat.get_chords()[0].number_of_dots == 3

    # def test_beat_update_chord_tuplets(self):
    #     self.fail()


class TestBeatUpdateChordBeams(TestCase):
    pass


class TestBeatAddChild(TestCase):
    def test_beat_add_child(self):
        v = create_voice()
        beats = v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        assert v.get_current_beat() == beats[0]
        chord1 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 8))[0]
        assert chord1.up == beats[0]
        assert v.get_current_beat() == beats[0]
        chord2 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 8))[0]
        assert chord2.up == beats[0]
        assert v.get_current_beat() == beats[1]
        chord3 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 4))[0]
        assert chord3.up == beats[1]
        assert v.get_current_beat() == beats[2]
        chord4 = v.get_current_beat().add_child(Chord(60, quarter_duration=1 / 2))[0]
        assert chord4.up == beats[2]
        with self.assertRaises(VoiceIsFullError):
            v.get_current_beat()
        assert beats[3].is_filled
        assert v.leftover_chord is None

        v = create_voice()
        beats = v.update_beats(1 / 4, 1 / 4, 1 / 4, 1 / 4)
        assert v.get_current_beat() == beats[0]
        v.get_current_beat().add_child(Chord(midis=60, quarter_duration=1 / 8))
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=1 / 4))
        assert v.get_current_beat() == beats[1]
        assert [ch.quarter_duration for ch in beats[0].get_children()] == [1 / 8, 1 / 8]
        assert [ch.midis[0].value for ch in beats[0].get_children()] == [60, 61]
        assert [ch.quarter_duration for ch in beats[1].get_children()] == [1 / 8]
        assert [ch.midis[0].value for ch in beats[1].get_children()] == [61]

    def test_add_child_with_fractional_quarter_duration_1(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=0.5))
        v.get_current_beat().add_child(Chord(midis=62, quarter_duration=3.5))
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [0.5, 0.5]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61, 62]
                assert beat.is_filled
            elif index == 1:
                assert [ch.quarter_duration for ch in beat.get_children()] == [3]
                assert [ch.midis[0].value for ch in beat.get_children()] == [62]
                assert beat.is_filled
            else:
                assert beat.get_children() == []
                assert beat.is_filled

    def test_add_child_with_fractional_quarter_duration_2(self):
        v = create_voice()
        beats = v.update_beats(1, 1.5, 1)
        v.get_current_beat().add_child(
            Chord(midis=61, quarter_duration=QuarterDuration(2, 5))
        )
        v.get_current_beat().add_child(Chord(midis=62, quarter_duration=4))
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [
                    QuarterDuration(2, 5),
                    QuarterDuration(3, 5),
                ]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61, 62]
                assert beat.is_filled
            elif index == 1:
                assert [ch.quarter_duration for ch in beat.get_children()] == [1.5]
                assert [ch.midis[0].value for ch in beat.get_children()] == [62]
                assert beat.is_filled
            elif index == 2:
                assert [ch.quarter_duration for ch in beat.get_children()] == [1]
                assert [ch.midis[0].value for ch in beat.get_children()] == [62]
                assert beat.is_filled
        assert v.leftover_chord.quarter_duration == 4 - 2.5 - 0.6


class TestBeatSplitChord(TestCase):
    def test_split_chord_different_accidental_instances(self):
        v1 = create_voice()
        v1.update_beats(1, 1.5)
        chord = Chord(midis=61, quarter_duration=2.5)
        v1.get_current_beat().add_child(chord)
        ch1, ch2 = v1.get_chords()

        assert ch1.midis[0].accidental != ch2.midis[0].accidental

    def test_split_chord_accidentals(self):
        v1 = create_voice()
        v1.update_beats(1, 1.5, 1)
        chord = Chord(midis=61, quarter_duration=7)
        chord.midis[0].accidental.show = True
        assert chord.midis[0].accidental.xml_object.value_ == "sharp"
        v1.get_current_beat().add_child(chord)
        v2 = create_voice()
        v2.update_beats(1, 1, 1)
        v2.get_current_beat().add_child(v1.leftover_chord)
        all_chords = v1.get_chords() + v2.get_chords()
        v1.up.up._update_divisions()
        v2.up.up._update_divisions()
        for b in v1.get_children() + v2.get_children():
            b.finalize()

        assert [
            ch.midis[0].accidental.xml_object.value_
            if ch.midis[0].accidental.xml_object
            else None
            for ch in all_chords
        ] == ["sharp", None, None, None]
        all_chords[1].midis[0].accidental.show = True

        assert [
            ch.midis[0].accidental.xml_object.value_
            if ch.midis[0].accidental.xml_object
            else None
            for ch in all_chords
        ] == ["sharp", "sharp", None, None]

    def test_split_not_writable_chords(self):
        v = create_voice()
        v.update_beats(1)
        v._add_chord(Chord(60, 5 / 6))
        v._add_chord(Chord(60, 1 / 6))
        assert [ch.quarter_duration for ch in v.get_chords()] == [5 / 6, 1 / 6]
        v.get_beat(1)._split_not_writable_chords()
        assert [ch.quarter_duration for ch in v.get_chords()] == [1 / 2, 1 / 3, 1 / 6]

    def test_add_child_5_leftover(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        split_children = v.get_current_beat().add_child(
            Chord(midis=61, quarter_duration=6)
        )
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [4]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61]
                assert beat.is_filled
            else:
                assert beat.get_children() == []
                assert beat.is_filled
        assert split_children[0].midis[0].is_tied_to_next
        assert v.leftover_chord.quarter_duration == 2
        assert not v.leftover_chord.split
        assert v.leftover_chord.midis[0].value == 61
        assert v.leftover_chord.midis[0].is_tied_to_previous

    def test_beat_add_child_1_3_no_split(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        v.get_current_beat().add_child(Chord(midis=60, quarter_duration=1))
        v.get_current_beat().add_child(Chord(midis=61, quarter_duration=3))
        assert v.leftover_chord is None
        for index, beat in enumerate(beats):
            if index == 0:
                assert [ch.quarter_duration for ch in beat.get_children()] == [1]
                assert [ch.midis[0].value for ch in beat.get_children()] == [60]
                assert beat.is_filled
            elif index == 1:
                assert [ch.quarter_duration for ch in beat.get_children()] == [3]
                assert [ch.midis[0].value for ch in beat.get_children()] == [61]
                assert beat.is_filled
            else:
                assert beat.get_children() == []
                assert beat.is_filled

    def test_beat_add_child_4_no_split(self):
        v = create_voice()
        beats = v.update_beats(1, 1, 1, 1)
        chord = v.get_current_beat().add_child(Chord(60, quarter_duration=4))
        assert v.leftover_chord is None
        for index, beat in enumerate(beats):
            if index == 0:
                assert beat.get_children() == chord
                assert beat.get_children()[0].quarter_duration == 4
                assert beat.is_filled
            else:
                assert beat.is_filled

    def test_split_tied_chords(self):
        v = create_voice()
        v.update_beats(1, 1, 1, 1)
        ch1 = Chord(60, quarter_duration=1.5)
        ch1.add_tie("start")
        split_chords_1 = v.get_current_beat().add_child(ch1)
        assert split_chords_1[-1].midis[0].is_tied_to_next

        ch2 = Chord([60, 63], quarter_duration=2)
        ch2.midis[1].add_tie("start")
        split_chords_2 = v.get_current_beat().add_child(ch2)
        assert not split_chords_2[-1].midis[0].is_tied_to_next
        assert split_chords_2[-1].midis[1].is_tied_to_next

    def test_add_chord_exception(self):
        b = Beat()
        with self.assertRaises(AddChordError):
            b.add_chord()

    def test_convert_splittables_dictionary(self):
        SPLITTABLES = {
            (0, 1): {
                (5, 6): [(3, 6), (2, 6)],
            },
            (1, 6): {
                (4, 6): [(2, 6), (2, 6)],
                (5, 6): [(2, 6), (3, 6)],
            },
            (2, 6): {
                (3, 6): [(2, 6), (1, 6)],
                (5, 6): [(2, 6), (3, 6)],
            },
        }

        expected = {
            QuarterDuration(0, 1): {
                QuarterDuration(5, 6): [QuarterDuration(3, 6), QuarterDuration(2, 6)],
            },
            QuarterDuration(1, 6): {
                QuarterDuration(4, 6): [QuarterDuration(2, 6), QuarterDuration(2, 6)],
                QuarterDuration(5, 6): [QuarterDuration(2, 6), QuarterDuration(3, 6)],
            },
            QuarterDuration(2, 6): {
                QuarterDuration(3, 6): [QuarterDuration(2, 6), QuarterDuration(1, 6)],
                QuarterDuration(5, 6): [QuarterDuration(2, 6), QuarterDuration(3, 6)],
            },
        }
        assert expected == _convert_to_quarter_duration_splittables_dictionary(
            SPLITTABLES
        )

    def test_all_splits(self):
        for key, value in SPLITTABLES.items():
            for k, v in value.items():
                p = Part(f"p{uuid.uuid4()}")
                qds = [QuarterDuration(*key), QuarterDuration(*k)]
                remaining_qd = QuarterDuration(1 - sum(qds))
                if remaining_qd < 0:
                    remaining_qd += 1
                if remaining_qd > 0:
                    qds.append(remaining_qd)
                for qd in qds:
                    if qd:
                        p.add_chord(Chord(70, qd))
                p.finalize()
                assert (
                    sum(
                        [
                            ch.quarter_duration
                            for ch in p.get_beats()[0].get_chords()
                            if not ch.is_rest
                        ]
                    )
                    == 1
                )


class TestNotImplementedTuplets(TestCase):
    def test_writing_subdivision_17(self):
        factors = [1, 2, 3, 4, 4, 3]
        qds = [QuarterDuration(x, 17) for x in factors]
        chords = [Chord(60, qd) for qd in qds]
        [ch.add_lyric(x) for ch, x in zip(chords, factors)]
        p = Part("p1")
        [p.add_chord(ch) for ch in chords]
        with self.assertRaises(QuarterDurationIsNotWritable):
            p.finalize()
        for ch in chords:
            if ch.quarter_duration == QuarterDuration(1, 17):
                ch.type = "64th"
                ch.number_of_dots = 0
            elif ch.quarter_duration == QuarterDuration(2, 17):
                ch.type = "32nd"
                ch.number_of_dots = 0
            elif ch.quarter_duration == QuarterDuration(3, 17):
                ch.type = "32nd"
                ch.number_of_dots = 1
            elif ch.quarter_duration == QuarterDuration(4, 17):
                ch.type = "16th"
                ch.number_of_dots = 0

            ch.tuplet = Tuplet(17, 16, "64th")

        p = Part("p2")
        [p.add_chord(ch) for ch in chords]
        p.finalize()
        beat = p.get_beats()[0]
        for ch in beat.get_chords():
            ch.check_printed_duration()
            ch.check_number_of_beams()



from typing import Optional

from musicscore.musictree import MusicTree
from musicscore.exceptions import (
    NoteHasNoParentChordError,
    NoteMidiHasNoParentChordError,
)
from musicscore.midi import Midi
from musicscore.quarterduration import QuarterDurationMixin
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import (
    XMLNote,
    XMLDot,
    XMLGrace,
    XMLRest,
    XMLTie,
    XMLNotations,
    XMLTied,
    XMLBeam,
)

__all__ = ["Note", "tie", "untie"]


def tie(*notes):
    """
    Tie notes.
    """
    if hasattr(notes[0], "__iter__"):
        notes = notes[0]
    notes[0].start_tie()
    if len(notes) > 1:
        for note in notes[1:-1]:
            note.stop_tie()
            note.start_tie()
        notes[-1].stop_tie()


def untie(*notes):
    """
    Untie notes.
    """
    if hasattr(notes[0], "__iter__"):
        notes = notes[0]
    notes[0].remove_tie("start")
    if len(notes) > 1:
        for note in notes[1:-1]:
            note.remove_tie("stop")
            note.remove_tie("start")
        notes[-1].remove_tie("stop")


class Note(MusicTree, XMLWrapper, QuarterDurationMixin):
    """
    Parent type: :obj:`~musicscore.chord.Chord`

    Child type: :obj:`~musicscore.midi.Midi`
    """

    _ATTRIBUTES = {
        "midi",
        "quarter_duration",
        "parent_chord",
        "is_tied",
        "is_tied_to_previous",
    }

    XMLClass = XMLNote

    def __init__(self, midi, quarter_duration=None, *args, **kwargs):
        self._midi = None
        self._parent_chord = midi.parent_chord
        self._xml_object = self.XMLClass(*args, **kwargs)

        super().__init__(quarter_duration=quarter_duration)
        self.midi = midi
        self._parent = self.parent_chord
        self._update_xml_notehead()
        self._update_xml_voice()
        self._update_xml_staff()
        self._update_xml_dots()
        self._update_xml_time_modification()
        self._update_xml_tuplet()
        self._update_xml_beams()

    @staticmethod
    def _check_xml_duration_value(duration):
        if int(duration) != duration:
            raise ValueError(
                f"product of quarter_duration and divisions {duration} must be an integer"
            )
        if duration < 0:
            raise ValueError

    def _add_child(self, child: Midi) -> Midi:
        """
        Check and add child to list of children. Child's parent is set to self.

        :param child: :obj:`~musicscore.midi.Midi`
        :return: child
        :rtype: :obj:`~musicscore.midi.Midi`
        """
        return super().add_child(child)

    def _set_xml_tied(self, val):
        if not self.xml_notations:
            self.xml_notations = XMLNotations()
        tied_xml_objects = self.xml_notations.find_children("XMLTied")
        tied_xml_types = [t.type for t in tied_xml_objects]
        if val in tied_xml_types:
            pass
        elif val == "stop" and "start" in tied_xml_types:
            tied_xml_objects[0].type = "stop"
            self.xml_notations.add_child(XMLTied(type="start"))
        else:
            self.xml_notations.add_child(XMLTied(type=val))

    def _update_ties(self):
        if "stop" in self.midi._ties:
            self.stop_tie()
        else:
            self.remove_tie("stop")
        if "start" in self.midi._ties:
            self.start_tie()
        else:
            self.remove_tie("start")

    def _update_xml_accidental(self):
        self.xml_object.xml_accidental = self.midi.accidental.xml_object

    def _update_xml_beams(self):
        if self.parent_chord and self.parent_chord.beams is not None:
            for number, value in self.parent_chord.beams.items():
                if value in ["forward", "backward"]:
                    value += " hook"
                self.xml_object.add_child(XMLBeam(number=number, value_=value))

    def _update_xml_dots(self):
        number_of_dots = self.parent_chord.number_of_dots
        if number_of_dots is None:
            number_of_dots = 0
        dots = self.xml_object.find_children("XMLDot")
        if number_of_dots > len(dots):
            diff = number_of_dots - len(dots)
            while diff:
                self.xml_object.add_child(XMLDot())
                diff -= 1
        elif number_of_dots < len(dots):
            for dot in dots[number_of_dots:]:
                dot.get_parent().remove(dot)
        else:
            pass

    def _update_xml_duration(self):
        duration = (
            float(self.quarter_duration) * self.get_parent_measure().get_divisions()
        )
        self._check_xml_duration_value(duration)
        duration = int(duration)
        if duration == 0:
            if self.midi and self.midi.value == 0:
                raise ValueError("A rest cannot be a grace note.")
            self.xml_object.xml_duration = None
            if not self.xml_object.xml_grace:
                self.xml_object.xml_grace = XMLGrace()
        else:
            self.xml_object.xml_grace = None
            self.xml_object.xml_duration = duration

    def _update_xml_pitch_or_rest(self):
        if self.midi.value == 0 and self.quarter_duration == 0:
            raise ValueError("A rest cannot be a grace note.")
        pitch_or_rest = self.midi.get_pitch_or_rest()
        if isinstance(pitch_or_rest, XMLRest):
            if self.xml_object.xml_pitch:
                self.xml_object.xml_pitch = None
            self.xml_object.xml_rest = pitch_or_rest
            self.xml_object.xml_notehead = None
        else:
            if self.xml_object.xml_rest:
                self.xml_object.xml_rest = None
            self.xml_object.xml_pitch = pitch_or_rest

    def _update_xml_staff(self):
        self.xml_object.xml_staff = self.get_staff_number()

    def _update_xml_type(self):
        self.xml_type = self.parent_chord.type

    def _update_xml_voice(self):
        self.xml_object.xml_voice = str(self.get_voice_number())

    def _update_xml_notations(self):
        """
        If ``self.xml_object.xml_notations`` has children of types :obj:`~musicxml.xmlelement.xmlelement.XMLArticulation` oder
        :obj:`~musicxml.xmlelement.xmlelement.XMLTechnical`, :obj:`~musicxml.xmlelement.xmlelement.XMLOrnaments`,
        :obj:`~musicxml.xmlelement.xmlelement.XMLDynamics`
        which have no children themselves, these will be removed.
        ``self.xml_object.xml_notations`` will be removed itself if it has no children.

        :return: None
        """
        if self.xml_object.xml_notations:
            if (
                self.xml_object.xml_notations.xml_articulations
                and not self.xml_object.xml_notations.xml_articulations.get_children()
            ):
                self.xml_object.xml_notations.remove(
                    self.xml_object.xml_notations.xml_articulations
                )

            if (
                self.xml_object.xml_notations.xml_technical
                and not self.xml_object.xml_notations.xml_technical.get_children()
            ):
                self.xml_object.xml_notations.remove(
                    self.xml_object.xml_notations.xml_technical
                )

            if (
                self.xml_object.xml_notations.xml_ornaments
                and not self.xml_object.xml_notations.xml_ornaments.get_children()
            ):
                self.xml_object.xml_notations.remove(
                    self.xml_object.xml_notations.xml_ornaments
                )

            if (
                self.xml_object.xml_notations.xml_dynamics
                and not self.xml_object.xml_notations.xml_dynamics.get_children()
            ):
                self.xml_object.xml_notations.remove(
                    self.xml_object.xml_notations.xml_dynamics
                )

            if not self.xml_object.xml_notations.get_children():
                self.xml_object.remove(self.xml_object.xml_notations)

    def _update_xml_notehead(self):
        self.xml_object.xml_notehead = self.midi.notehead

    def _update_xml_time_modification(self):
        if self.parent_chord.tuplet:
            self.xml_time_modification = (
                self.parent_chord.tuplet.get_xml_time_modification()
            )

    def _update_xml_tuplet(self):
        try:
            xml_tuplet = self.parent_chord.tuplet.get_xml_tuplet()
            if xml_tuplet:
                if not self.xml_notations:
                    self.xml_notations = XMLNotations()
                self.xml_notations.xml_tuplet = xml_tuplet
        except AttributeError:
            pass

    @property
    def is_tied(self) -> bool:
        """
        :return: True if an element :obj:`~musicxml.xmlelement.xmlelement.XMLTie` with type 'start' is under note's xml_object children.
        :rtype: bool
        """
        type_types = [t.type for t in self.xml_object.find_children("XMLTie")]
        if "start" in type_types:
            return True
        else:
            return False

    @property
    def is_tied_to_next(self) -> bool:
        """
        :return: same as :obj:`~musicscore.musicscore.Note.is_tied`
        """
        return self.is_tied

    @property
    def is_tied_to_previous(self) -> bool:
        """
        :return: True if an element :obj:`~musicxml.xmlelement.xmlelement.XMLTie` with type 'stop' is under note's xml_object children.
        :rtype: bool
        """
        type_types = [t.type for t in self.find_children("XMLTie")]
        if "stop" in type_types:
            return True
        else:
            return False

    @property
    def midi(self) -> Midi:
        """
        :obj:`~musicscore.chord.Chord`.midi property must be a :obj:`~musicscore.midi.Midi` object with a parent :obj:`~musicscore.chord.Chord`. :obj:`~musicscore.midi.Midi` with value `0` means rest.
        Setting this property will set :obj:`~musicscore.midi.Midi`\s :obj:`~musicscore.midi.Midi.parent_note` to self.

        :return: note's :obj:`~musicscore.midi.Midi`.
        """
        return self._midi

    @midi.setter
    def midi(self, value):
        if not isinstance(value, Midi):
            raise TypeError("Note.midi property must be of type Midi")
        if not value.parent_chord:
            raise NoteMidiHasNoParentChordError
        self._midi = value
        self._midi.parent_note = self
        self._update_xml_pitch_or_rest()
        self._update_xml_accidental()

    @property
    def parent_chord(self) -> "Chord":
        """
        :return: notes parent. Same as self.up
        """
        return self._parent_chord

    @QuarterDurationMixin.quarter_duration.setter
    def quarter_duration(self, value):
        if value is not None:
            if not self.parent_chord:
                raise NoteHasNoParentChordError()
            self._set_quarter_duration(value)
            self._update_xml_duration()
            self._update_xml_type()
            # self._update_xml_dots()
        else:
            self.xml_object.xml_duration = None

    def get_parent_chord(self):
        """
        returns :obj:`~parent_chord`
        """
        return self.parent_chord

    def get_or_create_xml_notations(self) -> "XMLNotations":
        """
        If note's ``xml_object`` has no :obj:`~musicxml.xmlelement.xmlelement.XMLNotations` as child this child will be created.

        :return: :obj:`~musicxml.xmlelement.xmlelement.XMLNotations`
        """
        if not self.xml_object.xml_notations:
            self.xml_object.xml_notations = XMLNotations()
        return self.xml_object.xml_notations

    def get_parent_measure(self) -> "Measure":
        """
        :return: parent :obj:`~musicscore.measure.Measure`
        """
        return self.parent_chord.get_parent_measure()

    def get_staff_number(self) -> int:
        """
        :return: number of parent :obj:`~musicscore.staff.Staff`
        """
        midi_staff_number = self.midi.get_staff_number()
        if midi_staff_number:
            return midi_staff_number

        return self.parent_chord.get_staff_number()

    def get_voice_number(self) -> int:
        """
        :return: number of parent :obj:`~musicscore.voice.Voice`
        """
        return self.get_parent_chord().get_voice_number()

    def remove_tie(self, type: Optional[str] = None) -> None:
        """
        :param type: 'start', 'stop', None: if None and note has :obj:`~musicxml.xmlelement.xmlelement.XMLTie` objects with both types
                      ValueError is raised.
        :return: None
        """
        ties = self.find_children("XMLTie")
        tie_to_be_removed = None
        if len(ties) == 0:
            pass
        elif len(ties) == 1:
            if type is None:
                tie_to_be_removed = ties[0]
            else:
                tie_to_be_removed = ties[0] if ties[0].type == type else None
        elif len(ties) == 2:
            if type is None:
                raise ValueError(
                    "Note has stop and start ties. Specify type=start or type=stop to decide which one should be removed"
                )
            else:
                tie_to_be_removed = [t for t in ties if t.type == type]
                tie_to_be_removed = (
                    None if not tie_to_be_removed else tie_to_be_removed[0]
                )
        else:
            raise NotImplementedError
        if tie_to_be_removed:
            try:
                self.midi._ties.remove(tie_to_be_removed.type)
            except KeyError:
                pass
            tied_to_be_removed = [
                t
                for t in self.xml_notations.find_children("XMLTied")
                if t.type == tie_to_be_removed.type
            ][0]
            tie_to_be_removed.up.remove(tie_to_be_removed)
            xml_notations = tied_to_be_removed.up
            xml_notations.remove(tied_to_be_removed)
            if not xml_notations.get_children():
                xml_notations.up.remove(xml_notations)

    def start_tie(self) -> None:
        """
        Start a tie if not already started.
        """
        if not self.is_tied:
            self.xml_object.add_child(XMLTie(type="start"))
            self._set_xml_tied("start")
        self.midi._ties.add("start")

    def stop_tie(self) -> None:
        """
        Stop a tie if not already stopped.
        """
        if self.is_tied_to_previous:
            pass
        elif self.is_tied:
            self.find_children("XMLTie")[0].type = "stop"
            self.xml_object.add_child(XMLTie(type="start"))
            self._set_xml_tied("stop")
        else:
            self.xml_object.add_child(XMLTie(type="stop"))
            self._set_xml_tied("stop")
        self.midi._ties.add("stop")



"""
Two voices share one staff. Each voice is assigned some lyrics.
"""

from pathlib import Path

from musicscore import Score, A, Chord, Lyrics, C, E, D, B, G
from musicscore.chord import Rest
from unittest import TestCase
from musicscore.util import slur_chords
from musicxml import XMLFermata, XMLAccent


class TestLily42a(TestCase):
    def test_lily_42a_MultiVoice_TwoVoicesOnStaff_Lyrics(self):
        score = Score()
        part = score.add_part("p1")
        v1 = [
            Chord(E(5), 2),
            Chord(D(5), 1),
            Chord(B(4), 1),
            Rest(1),
            Chord(D(5), 1),
            Chord(B(3), 1.5),
            Chord(C(5), 0.5),
            Rest(4),
        ]
        v2 = [
            Chord(C(5), 2),
            Chord(B(4), 1),
            Chord(G(4), 1),
            Rest(1),
            Chord(B(4), 1),
            Chord(G(3), 1.5),
            Chord(A(4), 0.5),
            Rest(4),
        ]
        for ch in [v1[0], v2[0]]:
            ch.add_x(XMLFermata())
            ch.add_x(XMLAccent())
        v1[1].add_dynamics("mf")
        slur_chords(v1[-3:-1])
        slur_chords(v2[-3:-1])
        [part.add_chord(ch, voice_number=1) for ch in v1]
        [part.add_chord(ch, voice_number=2) for ch in v2]
        Lyrics(
            ["This", "is", "the", None, "lyrics", "of", "Voice1", None],
            show_number=True,
            default_y=-110,
        ).add_to_chords(v1)
        Lyrics(
            ["This", "is", "the", None, "lyrics", "of", "Voice2", None],
            show_number=True,
            default_y=-130,
        ).add_to_chords(v2)
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from typing import Union, Optional

from musicscore.util import isinstance_as_string
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import (
    XMLPageLayout,
    XMLPageMargins,
    XMLScaling,
    XMLDefaults,
    XMLSystemLayout,
    XMLSystemMargins,
    XMLStaffLayout,
)

__all__ = [
    "PAGE_MARGINS",
    "PAGE_SIZES",
    "SYSTEM_MARGINS",
    "SYSTEM_LAYOUT",
    "STAFF_LAYOUT",
    "SCALING",
    "Margins",
    "Scaling",
    "PageLayout",
    "SystemLayout",
    "StaffLayout",
]
#:
PAGE_MARGINS = {
    "A4": {
        "portrait": {"left": 140, "right": 70, "top": 70, "bottom": 70},
        "landscape": {"left": 111, "right": 70, "top": 70, "bottom": 70},
    },
    "A3": {
        "portrait": {"left": 111, "right": 70, "top": 70, "bottom": 70},
        "landscape": {"left": 111, "right": 70, "top": 70, "bottom": 70},
    },
}

#:
PAGE_SIZES = {"A4": (209.991, 297.0389), "A3": (297.0389, 419.9819)}

#:
SYSTEM_MARGINS = {"left": 0, "right": 0}

#:
SYSTEM_LAYOUT = {"system_distance": 117, "top_system_distance": 117}

#:
STAFF_LAYOUT = {"staff_distance": 80}

#:
SCALING = {"millimeters": 7.2319, "tenths": 40}


class LayoutMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent = None

    @property
    def parent(self):
        """
        :return: musicscore object which uses this layout object.

        .. todo::
           Print implementation: At this moment only :obj:`~musicscore.score.Score` is implemented.
        """
        return self._parent

    @parent.setter
    def parent(self, val):
        self._parent = val
        if isinstance(self, PageLayout):
            if isinstance_as_string(self.parent, "Score"):
                self._update()
                self.parent.xml_object.xml_defaults.xml_page_layout = self.xml_object
            else:
                raise NotImplementedError
        elif isinstance_as_string(self, "SystemLayout"):
            if isinstance_as_string(self.parent, "Score"):
                self.parent.xml_object.xml_defaults.xml_system_layout = self.xml_object
            else:
                raise NotImplementedError
        elif isinstance_as_string(self, "StaffLayout"):
            if isinstance_as_string(self.parent, "Score"):
                self.parent.xml_object.xml_defaults.xml_staff_layout = self.xml_object
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError


class Margins:
    def __init__(
        self,
        parent: Union["PageLayout", "SystemLayout"],
        left: Union[float, int] = None,
        right: Union[float, int] = None,
        top: Union[float, int] = None,
        bottom: Union[float, int] = None,
    ):
        self._parent = None
        self._left = None
        self._right = None
        self._top = None
        self._parent_xml_object = None
        self._parent = None

        self.parent = parent
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

    def _update_parent(self, side):
        if side not in ["left", "right", "top", "bottom"]:
            raise ValueError
        if side in ["top", "bottom"] and isinstance(self.parent, SystemLayout):
            if eval(f"self.{side}") is not None:
                raise ValueError
        else:
            setattr(self._parent_xml_object, f"xml_{side}_margin", eval(f"self.{side}"))

    @property
    def bottom(self):
        """
        Set and get :obj:`~musicxml.xmlelement.xmlelement.XMLBottomMargin` value of parent.
        """
        return self._bottom

    @bottom.setter
    def bottom(self, val):
        self._bottom = val
        self._update_parent("bottom")

    @property
    def left(self) -> Union[int, float]:
        """
        Set and get :obj:`~musicxml.xmlelement.xmlelement.XMLLeftMargin` value of parent.
        """
        return self._left

    @left.setter
    def left(self, val):
        self._left = val
        self._update_parent("left")

    @property
    def parent(self):
        """
        :return: parent layout object. :obj:`PageLayout` and :obj:`SystemLayout` are implemented.
        """
        return self._parent

    @parent.setter
    def parent(self, val):
        self._parent = val
        if isinstance(val, PageLayout):
            self._parent_xml_object = self._parent.xml_object.xml_page_margins
        elif isinstance(val, SystemLayout):
            self._parent_xml_object = self._parent.xml_object.xml_system_margins
        else:
            raise NotImplementedError

    @property
    def right(self) -> Union[int, float]:
        """
        Set and get :obj:`~musicxml.xmlelement.xmlelement.XMLRightMargin` value of parent.
        """
        return self._right

    @right.setter
    def right(self, val):
        self._right = val
        self._update_parent("right")

    @property
    def top(self) -> Union[int, float]:
        """
        Set and get :obj:`~musicxml.xmlelement.xmlelement.XMLTopMargin` value of parent.
        """
        return self._top

    @top.setter
    def top(self, val) -> Union[int, float]:
        self._top = val
        self._update_parent("top")


class Scaling(XMLWrapper):
    _ATTRIBUTES = {"millimeters", "tenths", "score"}
    XMLClass = XMLScaling

    def __init__(
        self,
        millimeters: Union[int, float] = SCALING["millimeters"],
        tenths: Union[int, float] = SCALING["tenths"],
    ):
        super().__init__()
        self._xml_object = self.XMLClass()
        self._millimeters = None
        self._tenths = None
        self._score = None

        self.millimeters = millimeters
        self.tenths = tenths

    def _update_score(self):
        if self.score:
            self.score.page_layout._set_page_height_and_width()

    @property
    def millimeters(self) -> Union[int, float]:
        """
        Set and get millimeters value of scaling object. After setting value, parent :obj:`~musicscore.score.Score`'s :obj:`PageLayout` is
        updated to reflect the changes.

        :return: millimeters
        :rtype: Union[int, float]
        """
        return self._millimeters

    @millimeters.setter
    def millimeters(self, val):
        if val != self._millimeters:
            self._millimeters = val
            self.xml_object.xml_millimeters = val
            self._update_score()

    @property
    def score(self):
        """
        Set and get parent :obj:`~musicscore.score.Score`. After setting score, its :obj:`~musicxml.xmlelement.xmlelement.XMLScaling`
        and :obj:`~musicxml.xmlelement.xmlelement.XMLDefaults` are created if needed.

        :return: parent score
        :rtype: :obj:`~musicscore.score.Score`
        """
        return self._score

    @score.setter
    def score(self, val):
        self._score = val
        if not self.score.xml_object.xml_defaults:
            self.score.xml_object.xml_defaults = XMLDefaults()
        self.score.xml_object.xml_defaults.xml_scaling = self.xml_object

    @property
    def tenths(self) -> Union[int, float]:
        """
        Set and get tenths value of scaling object. After setting value, parent :obj:`~musicscore.score.Score`'s :obj:`PageLayout` is updated to reflect the
        changes.

        :return: tenths
        :rtype: Union[int, float]
        """
        return self._tenths

    @tenths.setter
    def tenths(self, val):
        if val != self._tenths:
            self._tenths = val
            self.xml_object.xml_tenths = val
            self._update_score()

    def millimeters_to_tenths(self, x: Union[int, float]) -> Union[int, float]:
        """
        Converts millimeter value into tenths

        :param x: millimeters
        :return: calculated tenths
        """
        return round((x * self.tenths) / self.millimeters)


class PageLayout(XMLWrapper, LayoutMixin):
    """
    :param size: :obj:`PAGE_SIZES`
    :param orientation: 'portrait', 'landscape'
    """

    _ATTRIBUTES = {"scaling", "size", "orientation", "parent"}
    XMLClass = XMLPageLayout

    def __init__(self, size: str = "A4", orientation: str = "portrait"):
        super().__init__()
        self._xml_object = self.XMLClass()
        self._xml_object.xml_page_margins = XMLPageMargins(type="both")

        self._size = None
        self._orientation = None

        self.size = size
        self.orientation = orientation

        self._margins = Margins(
            parent=self, **PAGE_MARGINS[self.size][self.orientation]
        )

    def _get_page_height(self):
        return (
            self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][1])
            if self.orientation == "portrait"
            else self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][0])
        )

    def _get_page_width(self):
        return (
            self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][0])
            if self.orientation == "portrait"
            else self.scaling.millimeters_to_tenths(PAGE_SIZES[self.size][1])
        )

    def _set_page_height_and_width(self):
        self.xml_object.xml_page_height = self._get_page_height()
        self.xml_object.xml_page_width = self._get_page_width()

    def _update(self):
        self._set_page_height_and_width()
        self._margins = Margins(
            parent=self, **PAGE_MARGINS[self.size][self.orientation]
        )

    @property
    def margins(self) -> Margins:
        """
        Gets margins attribute.

        :return: margins object
        :rtype: :obj:`Margins`
        """
        return self._margins

    @property
    def orientation(self) -> str:
        """
        Set and get orientation. Permitted values are ['portrait', 'landscape']. After setting value, if parent and size already
        exist, page's height and width are set.

        :return: 'portrait', 'landscape'
        :rtype: str
        """
        return self._orientation

    @orientation.setter
    def orientation(self, val):
        _permitted = ["portrait", "landscape"]
        if val not in _permitted:
            raise ValueError(f"{val} can only be: {_permitted}")
        if val != self._orientation:
            self._orientation = val
            if self.size and self.parent:
                self._update()

    @property
    def scaling(self) -> Scaling:
        """
        :return: :obj:`~musicscore.score.Score`'s :obj:`Scaling`.
        :rtype: :obj:`Scaling`
        """
        return self.parent.get_root().scaling

    @property
    def size(self) -> str:
        """
        Set and get orientation. Permitted values are keys of obj:`PAGE_SIZES`. After setting value, if parent and orientation already
        exist, page's height and width are set.

        :return: sizes in :obj:`PAGE_SIZES`
        :rtype: str
        """
        return self._size

    @size.setter
    def size(self, val):
        if val not in PAGE_SIZES:
            raise NotImplementedError
        if val != self._size:
            self._size = val
            if self.orientation and self.parent:
                self._update()


class SystemLayout(XMLWrapper, LayoutMixin):
    _ATTRIBUTES = {"system_distance", "top_system_distance", "parent"}
    XMLClass = XMLSystemLayout

    def __init__(
        self,
        system_distance: Union[int, float] = SYSTEM_LAYOUT["system_distance"],
        top_system_distance: Union[int, float] = SYSTEM_LAYOUT["top_system_distance"],
    ):
        super().__init__()

        self._xml_object = self.XMLClass()
        self._xml_object.xml_system_margins = XMLSystemMargins()

        self.system_distance = system_distance
        self.top_system_distance = top_system_distance

        self._margins = Margins(parent=self, **SYSTEM_MARGINS)

    @property
    def margins(self) -> Margins:
        """
        Gets margins attribute.

        :return: margins object
        :rtype: :obj:`Margins`
        """
        return self._margins

    @property
    def system_distance(self) -> Optional[Union[int, float]]:
        """
        Set and get ``value`` of :obj:`~musicxml.xmlelement.xmlelement.XMLSystemDistance`.

        :return: ``self.xml_object.xml_system_distance.value_``
        :rtype: int, float, None
        """
        if self.xml_object.xml_system_distance:
            return self.xml_object.xml_system_distance.value_

    @system_distance.setter
    def system_distance(self, val):
        self.xml_object.xml_system_distance = val

    @property
    def top_system_distance(self) -> Optional[Union[int, float]]:
        """
        Set and get ``value`` of :obj:`~musicxml.xmlelement.xmlelement.XMLTopSystemDistance`.

        :return: ``self.xml_object.xml_top_system_distance.value_``
        :rtype: int, float, None
        """
        if self.xml_object.xml_top_system_distance:
            return self.xml_object.xml_top_system_distance.value_

    @top_system_distance.setter
    def top_system_distance(self, val):
        self.xml_object.xml_top_system_distance = val


class StaffLayout(XMLWrapper, LayoutMixin):
    _ATTRIBUTES = {"staff_distance", "parent"}
    XMLClass = XMLStaffLayout

    def __init__(self, staff_distance=STAFF_LAYOUT["staff_distance"]):
        super().__init__()
        self._xml_object = self.XMLClass()
        self.staff_distance = staff_distance

    @property
    def staff_distance(self) -> Optional[Union[int, float]]:
        """
        Set and get ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLStaffDistance`.

        :return: ``self.xml_object.xml_staff_distance.value_``
        :rtype: int, float, None
        """
        if self.xml_object.xml_staff_distance:
            return self.xml_object.xml_staff_distance.value_

    @staff_distance.setter
    def staff_distance(self, val):
        self.xml_object.xml_staff_distance = val



from typing import Optional, List

from musicscore.quarterduration import QuarterDuration
from musicscore.util import isinstance_as_string


class QuantizeMixin:
    _ATTRIBUTES = {"get_quantized"}

    def __init__(self, get_quantized=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._possible_subdivisions = {}
        self._get_quantized = None
        self.get_quantized = get_quantized

    def _get_beat_quarter_duration(self):
        if isinstance_as_string(self, "Beat"):
            beat_quarter_duration = self.quarter_duration
        else:
            beat_quarter_duration = QuarterDuration(1)
        return beat_quarter_duration

    @property
    def get_quantized(self) -> bool:
        """
        :obj:`~musicscore.quantize.QuantizeMixin` property

        - If ``get_quantized`` is set to None the first ``get_quantized`` of ancestors which is ``False`` or ``True`` will be returned.
        - If :obj:`musicscore.score.Score.get_quantized` is set to None it will be converted to ``False``
        - :obj:`musicscore.measure.Measure.finalize()` loops over all beats. If :obj:`musicscore.beat.Beat.get_quantized` returns ``True``
          :obj:`musicscore.beat.Beat.quantize_quarter_durations()` is called.

        :type: Optional[bool]
        :rtype: bool
        """
        if self._get_quantized is None:
            if self.up:
                return self.up.get_quantized
            else:
                return False
        return self._get_quantized

    @get_quantized.setter
    def get_quantized(self, val):
        self._get_quantized = val

    def get_possible_subdivisions(
        self, beat_quarter_duration: Optional[QuarterDuration] = None
    ) -> List[int]:
        """
        :obj:`~musicscore.quantize.QuantizeMixin` method

        This method is used by :obj:`~musicscore.beat.Beat`'s :obj:`~musicscore.beat.Beat.quantize_quarter_durations()`.

        Possible subdivisions dictionary can be set with :obj:`~musicscore.quantize.QuantizeMixin.set_possible_subdivisions()`. Keys in the subdivisions dictionary correspond to beat quarter durations (e.g. 1, 1/2 etc.)
        For example if get_possible_subdivisions()[1] ==  [3, 5, 8] a beat with a quarter duration of 1 can after quantization consists only of eighth triplets, quintuplets and 32ths.
        If this dictionary is not set or ``beat_quarter_duration`` as key does not exist, the parent's possible subdivisions dictionary will be checked.

        :obj:`~musicscore.score.Score` has a default :obj:`~musicscore.score.POSSIBLE_SUBDIVISIONS` dictionary which will be used if no other
        musicscore node on the path from self to root has its own possible subdivisions dictionary with ``beat_quarter_duration`` as a
        key. For setting possible subdivisions dictionary use always :obj:`~musicscore.quantize.QuantizeMixin.set_possible_subdivisions()`.

        :param beat_quarter_duration: Used as key in possible subdivisions dictionary.
               If ``None`` and self is a :obj:`~musicscore.beat.Beat` ``self.quarter_duration`` is used.
               If ``None`` and self is not a :obj:`~musicscore.beat.Beat` it is set to 1.
        :return: A list of possible subdivisions of a :obj:`~musicscore.beat.Beat`. This is used by beat's
                 :obj:`~musicscore.beat.Beat.quantize_quarter_durations()`
        :rtype: List[int]
        """
        if beat_quarter_duration is None:
            beat_quarter_duration = self._get_beat_quarter_duration()
        subdivisions = self._possible_subdivisions.get(beat_quarter_duration)
        if (
            subdivisions is None
            and self.up is not None
            and self.up.get_possible_subdivisions(beat_quarter_duration) is not None
        ):
            subdivisions = self.up.get_possible_subdivisions(beat_quarter_duration)[:]
        return subdivisions

    def set_possible_subdivisions(
        self,
        subdivisions: list[int],
        beat_quarter_duration: Optional[QuarterDuration] = None,
    ) -> None:
        """
        :obj:`~musicscore.quantize.QuantizeMixin` method

        This method is used to set or change possible subdivisions dictionary of a beat or its ascendants.
        For example if get_possible_subdivisions()[1] ==  [3, 5, 8] a beat with a quarter duration of 1 can after quantization consists only of eighth triplets, quintuplets and 32ths.

        :param subdivisions: list of possible subdivisions to be used in :obj:`musicscore.beat.Beat.quantize_quarter_durations()`
        :param beat_quarter_duration: If ``None`` and self is a :obj:`~musicscore.beat.Beat` ``self.quarter_duration`` is used.
                                      If ``None`` and self is not a :obj:`~musicscore.beat.Beat` it is set to 1.
        :return: None
        """
        if beat_quarter_duration is None:
            beat_quarter_duration = self._get_beat_quarter_duration()
        elif (
            isinstance_as_string(self, "Beat")
            and beat_quarter_duration != self.quarter_duration
        ):
            raise ValueError(
                f"beat_quarter_duration '{beat_quarter_duration}' must be None or equal to the beat quarter_duration '{self.quarter_duration}'"
            )

        if not isinstance(beat_quarter_duration, QuarterDuration):
            beat_quarter_duration = QuarterDuration(beat_quarter_duration)
        try:
            subdivisions = list(subdivisions)
        except TypeError:
            subdivisions = [subdivisions]
        self._possible_subdivisions[beat_quarter_duration] = subdivisions



from pathlib import Path

from musicscore import Score, Chord, Time
from musicscore.lyrics import Lyrics
from unittest import TestCase
from musicscore.util import slur_chords

"""Beaming or slurs can indicate melismata for lyrics. Also make sure that notes without an explicit syllable are
treated as if they were part of a melisma."""

"""
Chord.break_beam() will be needed. Otherwise breaking the beams like in this test will effect all staves and voices
"""


class TestLily61h(TestCase):
    def test_lily_61h_BeamsMelismata(self):
        score = Score()
        part = score.add_part("p1")
        t = Time(4, 4)
        t.actual_signatures = (3, 8, 1, 8, 3, 8, 1, 8)
        part.add_measure(time=t)
        t = Time(4, 4)
        t.actual_signatures = 8 * [1, 8]
        part.add_measure(time=t)

        midi_pattern = [72, 69, 72, 69, 72, 67, 71, 0]
        chords = []
        for _ in range(4):
            chords += [Chord(m, 0.5) for m in midi_pattern]

        [part.add_chord(ch) for ch in chords]

        third_measure_chords = part.get_measure(measure_number=3).get_chords()
        slur_chords(third_measure_chords[:3])
        slur_chords(third_measure_chords[4:7])

        fourth_measure_chords = part.get_measure(measure_number=4).get_chords()
        slur_chords(fourth_measure_chords[:7])

        lyrics_pattern = [("Me", None, None, "lis", "ma", None, None)]
        for m in part.get_children():
            Lyrics(lyrics_pattern).add_to_chords(m.get_chords()[:-1])

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Rests can have explicit pitches, where they are displayed. The first rest uses no explicit position and should use the default position, all others are explicitly positioned somewhere else.
"""

from pathlib import Path

from musicscore import Score
from musicscore.chord import Rest
from unittest import TestCase


class TestLily02b(TestCase):
    def test_lily_02b_Rests_PitchedRests(self):
        score = Score()
        part = score.add_part("p1")

        part.add_measure(time=(5, 4))
        rest_step_octaves = [None, ("D", 4), ("F", 5), ("A", 3), ("C", 6)]
        rests = [Rest(1, so[0], so[1]) if so else Rest(1) for so in rest_step_octaves]
        [part.add_chord(r) for r in rests]
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from musicscore.chord import Chord
from musicscore.dynamics import DYNAMICS, Dynamics
from musicxml.xmlelement.xmlelement import XMLPp, XMLFf, XMLSfpp, XMLP, XMLF

from musicscore.tests.util import ChordTestCase


class TestDynamics(ChordTestCase):
    def test_dynamics_init(self):
        d = Dynamics("pp")
        assert d.xml_object.__class__ == XMLPp

        d = Dynamics("sfpp")
        assert d.xml_object.__class__ == XMLSfpp

    def test_add_dynamics_to_a_chord_consequently(self):
        ch = Chord(60, 1)
        d1 = ch.add_dynamics("p")[0]
        d2 = ch.add_dynamics("f")[0]
        assert ch.xml_direction_types["below"] == [
            ("dynamics", [d1]),
            ("dynamics", [d2]),
        ]
        ch._parent = self.mock_beat
        ch.finalize()
        assert len(ch._xml_directions) == 2
        assert isinstance(
            ch._xml_directions[0].xml_direction_type.xml_dynamics.get_children()[0],
            XMLP,
        )
        assert isinstance(
            ch._xml_directions[1].xml_direction_type.xml_dynamics.get_children()[0],
            XMLF,
        )

    def test_add_dynamics(self):
        for dynamics in DYNAMICS:
            ch = Chord(60, 1)
            ch._parent = self.mock_beat
            ch.add_dynamics(dynamics)
            ch.finalize()
            assert len(ch._xml_directions) == 1
            d = ch._xml_directions[0]
            assert d.placement == "below"
            assert (
                d.xml_direction_type.xml_dynamics.get_children()[0].__class__
                == DYNAMICS[dynamics]
            )

    def test_add_dynamics_placement(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        ch.add_dynamics("pp", placement="above")
        ch.finalize()
        d = ch._xml_directions[0]
        assert d.placement == "above"

    def test_add_multiple_dynamics(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        ch.add_dynamics(["pp", "ff"], "below")
        ch.finalize()
        d = ch._xml_directions[0]
        dts = d.find_children("XMLDirectionType")
        assert d.placement == "below"
        assert dts[0].xml_dynamics.get_children()[0].__class__ == XMLPp
        assert dts[1].xml_dynamics.get_children()[0].__class__ == XMLFf

        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        ch.add_dynamics("pp", "above")
        ch.add_dynamics("ff", "below")
        ch.finalize()
        ds = ch._xml_directions
        assert [ds[0].placement, ds[1].placement] == ["above", "below"]
        assert (
            ds[0].xml_direction_type.xml_dynamics.get_children()[0].__class__ == XMLPp
        )
        assert (
            ds[1].xml_direction_type.xml_dynamics.get_children()[0].__class__ == XMLFf
        )



"""
Some tuples (3:2, 3:2, 3:2, 4:2, 4:1, 7:3, 6:2) with the default tuplet bracket displaying the number of actual
notes played. The second tuplet does not have a number attribute set.
"""

from pathlib import Path

from musicscore import Score, Time, Chord, A, B, C, D, E, F, G
from unittest import TestCase


class TestLily23a(TestCase):
    def test_lily_23a_Tuplets(self):
        score = Score()
        part = score.add_part("p1")

        part.add_measure(Time(2, 2))
        scale = [
            C(4),
            D(4),
            E(4),
            F(4),
            G(4),
            A(4),
            B(4),
            C(5),
            D(5),
            E(5),
            F(5),
            G(5),
            A(5),
            B(5),
            C(6),
        ]
        scale += list(reversed(scale))
        for m, d in zip(scale, 9 * [2 / 3] + 8 * [1] + 7 * [2 / 7] + 6 * [2 / 3]):
            part.add_chord(Chord(m, d))

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Some notes with simple lyrics: Syllables, notes without a syllable, syllable spanners.
"""

from pathlib import Path

from musicscore import Score, Chord, A
from musicscore.lyrics import Lyrics
from unittest import TestCase


class TestLily61a(TestCase):
    def test_lily_61a_Lyrics(self):
        score = Score()
        part = score.add_part("p1")

        chords = [Chord(A(4), 1) for _ in range(6)]
        chords.append(Chord(A(4), 2))

        lyrics = Lyrics([("Tra", "la", "la"), "Ja!", ("Tra", "ra!"), "Bah"])
        lyrics.add_to_chords(chords)

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from unittest import TestCase
from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score


class TestGetChords(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        p1 = self.score.add_child(Part("p1"))
        p2 = self.score.add_child(Part("p2"))
        p1.add_chord(Chord(60, 1))
        p1.add_chord(Chord(61, 3))
        p1.add_chord(Chord(62, 4))
        p1.add_chord(Chord(63, 2.5), voice_number=2)
        p1.add_chord(Chord(64, 1.5), voice_number=2)
        p2.add_chord(Chord(48, 8), staff_number=2, voice_number=2)

    def test_beat_get_chords(self):
        pass

    def test_voice_get_chords(self):
        v1111 = self.score.get_part(1).get_measure(1).get_staff(1).get_voice(1)
        v1112 = self.score.get_part(1).get_measure(1).get_staff(1).get_voice(2)
        # v1121 = None
        # v1122 = None
        v1211 = self.score.get_part(1).get_measure(2).get_staff(1).get_voice(1)
        v1212 = self.score.get_part(1).get_measure(2).get_staff(1).get_voice(2)
        # v1221 = None
        # v1222 = None
        v2111 = self.score.get_part(2).get_measure(1).get_staff(1).get_voice(1)
        v2112 = self.score.get_part(2).get_measure(1).get_staff(1).get_voice(2)
        v2121 = self.score.get_part(2).get_measure(1).get_staff(2).get_voice(1)
        v2122 = self.score.get_part(2).get_measure(1).get_staff(2).get_voice(2)
        v2211 = self.score.get_part(2).get_measure(2).get_staff(1).get_voice(1)
        v2212 = self.score.get_part(2).get_measure(2).get_staff(1).get_voice(2)
        v2221 = self.score.get_part(2).get_measure(2).get_staff(2).get_voice(1)
        v2222 = self.score.get_part(2).get_measure(2).get_staff(2).get_voice(2)

        chords = v1111.get_chords()
        assert [chord.midis[0].value for chord in chords] == [60, 61]
        assert [chord.quarter_duration for chord in chords] == [1, 3]

        chords = v1112.get_chords()
        assert [chord.midis[0].value for chord in chords] == [63, 63, 64, 64]
        assert [chord.quarter_duration for chord in chords] == [2, 0.5, 0.5, 1]

        chords = v1211.get_chords()
        assert [chord.midis[0].value for chord in chords] == [62]
        assert [chord.quarter_duration for chord in chords] == [4]

        assert v1212 is None

        chords = v2111.get_chords()
        assert [chord.midis[0].value for chord in chords] == []
        assert [chord.quarter_duration for chord in chords] == []

        assert v2112 is None

        chords = v2121.get_chords()
        assert [chord.midis[0].value for chord in chords] == []
        assert [chord.quarter_duration for chord in chords] == []

        chords = v2122.get_chords()
        assert [chord.midis[0].value for chord in chords] == [48]
        assert [chord.quarter_duration for chord in chords] == [4]

        chords = v2211.get_chords()
        assert [chord.midis[0].value for chord in chords] == []
        assert [chord.quarter_duration for chord in chords] == []

        assert v2212 is None

        chords = v2221.get_chords()
        assert [chord.midis[0].value for chord in chords] == []
        assert [chord.quarter_duration for chord in chords] == []

        chords = v2222.get_chords()
        assert [chord.midis[0].value for chord in chords] == [48]
        assert [chord.quarter_duration for chord in chords] == [4]

    def test_staff_get_chords(self):
        st111 = self.score.get_part(1).get_measure(1).get_staff(1)
        st121 = self.score.get_part(1).get_measure(2).get_staff(1)
        st211 = self.score.get_part(2).get_measure(1).get_staff(1)
        st221 = self.score.get_part(2).get_measure(2).get_staff(1)
        st212 = self.score.get_part(2).get_measure(1).get_staff(2)
        st222 = self.score.get_part(2).get_measure(2).get_staff(2)

        chords = st111.get_chords()
        assert [chord.midis[0].value for chord in chords] == [60, 61, 63, 63, 64, 64]
        assert [chord.quarter_duration for chord in chords] == [1, 3, 2, 0.5, 0.5, 1]

        chords = st121.get_chords()
        assert [chord.midis[0].value for chord in chords] == [62]
        assert [chord.quarter_duration for chord in chords] == [4]

        assert st211.get_chords() == []
        assert st221.get_chords() == []

        chords = st212.get_chords()
        assert [chord.midis[0].value for chord in chords] == [48]
        assert [chord.quarter_duration for chord in chords] == [4]

        chords = st222.get_chords()
        assert [chord.midis[0].value for chord in chords] == [48]
        assert [chord.quarter_duration for chord in chords] == [4]

    def test_part_get_chords(self):
        chords = self.score.get_part(1).get_chords()
        assert [chord.midis[0].value for chord in chords] == [
            60,
            61,
            63,
            63,
            64,
            64,
            62,
        ]
        assert [chord.quarter_duration for chord in chords] == [1, 3, 2, 0.5, 0.5, 1, 4]

    def test_score_get_chords(self):
        assert [ch.midis[0].value for ch in self.score.get_chords()] == [
            60,
            61,
            63,
            63,
            64,
            64,
            62,
            48,
            48,
        ]
        assert [ch.quarter_duration for ch in self.score.get_chords()] == [
            1,
            3,
            2,
            0.5,
            0.5,
            1,
            4,
            4,
            4,
        ]



"""
Text markup: different font sizes, weights and colors.
Bold, Medium
Bold, Large
Bold, Small
Normal, Medium
Normal, Large
Normal, Small
Normal, Small, Colored, Below
"""

from pathlib import Path

from musicscore import Score, Time, Chord, G, F
from unittest import TestCase


class TestLily32b(TestCase):
    def test_lily_32b_Articulations_Texts(self):
        score = Score()
        part = score.add_part("p1")
        t = Time(4, 4)
        t.actual_signatures = [1, 8]
        m = part.add_measure(t)
        m.implicit = "yes"
        m.number = 0
        part.add_measure(Time(4, 4))
        part.add_chord(Chord(F(4), 0.5))
        part.add_chord(Chord(G(4), 4))
        part.add_chord(Chord(F(4), 4))

        chords = part.get_chords()
        chords[0].add_words(
            "Bold, Medium", placement="below", font_weight="bold", font_size=14
        )
        chords[1].add_words(
            "Bold, Large",
            placement="below",
            font_weight="bold",
            font_size=18,
            relative_y=-30,
        )
        chords[2].add_words(
            "Bold, Small", placement="below", font_weight="bold", font_size=10
        )
        chords[0].add_words("Normal, Medium", font_size=18)
        chords[1].add_words("Normal, Large", font_size=14, relative_y=30)
        chords[2].add_words("Normal, Small", font_size=10)
        chords[2].add_words(
            "Normal, Small, Colored, Below",
            placement="below",
            font_size=10,
            color="#FF0000",
            relative_y=-30,
        )

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase
from musicscore.util import XML_DIRECTION_TYPE_CLASSES
from musicxml.xmlelement.xmlelement import (
    XMLSymbol,
    XMLDynamics,
    XMLWedge,
    XMLDashes,
    XMLBracket,
    XMLPedal,
    XMLMetronome,
    XMLBeatUnit,
    XMLPerMinute,
    XMLOctaveShift,
    XMLHarpPedals,
    XMLPedalTuning,
    XMLPedalStep,
    XMLPedalAlter,
    XMLStringMute,
    XMLScordatura,
    XMLAccord,
    XMLTuningOctave,
    XMLTuningStep,
    XMLImage,
    XMLPrincipalVoice,
    XMLPercussion,
    XMLWood,
    XMLStaffDivide,
)


class TestDynamics(TestCase):
    def test_direction_types(self):
        score = Score(title="Directions")
        p = score.add_part("part-1")
        for dt_class in [cl for cl in XML_DIRECTION_TYPE_CLASSES if cl != XMLDynamics]:
            if dt_class == XMLSymbol:
                dt_obj = dt_class("0")
            elif dt_class == XMLWedge:
                dt_obj = dt_class(type="crescendo")
            elif dt_class == XMLDashes:
                dt_obj = dt_class(type="start")
            elif dt_class == XMLBracket:
                dt_obj = dt_class(type="start", line_end="none")
            elif dt_class == XMLPedal:
                dt_obj = dt_class(type="start")
            elif dt_class == XMLMetronome:
                dt_obj = dt_class()
                dt_obj.add_child(XMLBeatUnit("quarter"))
                dt_obj.add_child(XMLPerMinute("120"))
            elif dt_class == XMLOctaveShift:
                dt_obj = dt_class(type="up")
            elif dt_class == XMLHarpPedals:
                dt_obj = dt_class()
                pt = dt_obj.add_child(XMLPedalTuning())
                pt.add_child(XMLPedalStep("A"))
                pt.add_child(XMLPedalAlter(1))
            elif dt_class == XMLStringMute:
                dt_obj = dt_class(type="on")
            elif dt_class == XMLScordatura:
                dt_obj = dt_class()
                acc = dt_obj.add_child(XMLAccord())
                acc.add_child(XMLTuningStep("A"))
                acc.add_child(XMLTuningOctave(0))
            elif dt_class == XMLImage:
                # dt_obj = dt_class(source='www.example.com')
                continue
            elif dt_class == XMLPrincipalVoice:
                dt_obj = dt_class(type="start", symbol="none")
            elif dt_class == XMLPercussion:
                dt_obj = dt_class()
                dt_obj.add_child(XMLWood("cabasa"))
            elif dt_class == XMLStaffDivide:
                dt_obj = dt_class(type="up")
            else:
                dt_obj = dt_class()

            chord = Chord(midis=60, quarter_duration=4)
            chord.add_direction_type(dt_obj)
            p.add_chord(chord)
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from musicscore.quarterduration import QuarterDuration
from musicscore.exceptions import (
    QuarterDurationIsNotWritable,
    MetronomeWrongBeatUnitError,
)
from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLMetronome, XMLBeatUnitDot, XMLSound


class Metronome(XMLWrapper):
    _ATTRIBUTES = {"per_minute", "beat_unit"}
    XMLClass = XMLMetronome

    def __init__(self, per_minute, beat_unit=1, parenthesis=False, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._per_minute = None
        self._beat_unit = None
        self._sound = XMLSound()

        self.per_minute = per_minute
        self.beat_unit = beat_unit

    def _set_xml_beat_unit(self):
        if self.beat_unit.denominator not in [1, 2, 4, 8, 16, 32]:
            raise MetronomeWrongBeatUnitError(self.beat_unit)
        try:
            self.xml_object.xml_beat_unit = self.beat_unit.get_type()
        except QuarterDurationIsNotWritable:
            raise MetronomeWrongBeatUnitError(self.beat_unit)
        self._set_xml_sound_tempo()

    def _set_xml_per_minute(self):
        self.xml_object.xml_per_minute = str(self.per_minute)
        self._set_xml_sound_tempo()

    def _set_xml_sound_tempo(self):
        if self.beat_unit:
            if not self._sound:
                self._sound = XMLSound()
            self._sound.tempo = int(self.beat_unit.value * self.per_minute)

    def _set_xml_beat_unit_dots(self):
        number_of_dots = self.beat_unit.get_number_of_dots()
        current_xml_unit_dots = [
            ch
            for ch in self.xml_object.get_children()
            if isinstance(ch, XMLBeatUnitDot)
        ]
        number_of_superfluous_dots = len(current_xml_unit_dots) - number_of_dots
        if number_of_superfluous_dots > 0:
            for xml_unit_dot in current_xml_unit_dots[:number_of_superfluous_dots]:
                self.xml_object.remove(xml_unit_dot)
        elif number_of_superfluous_dots < 0:
            for _ in range(number_of_superfluous_dots * -1):
                self.xml_object.add_child(XMLBeatUnitDot())
        else:
            pass

    @property
    def beat_unit(self):
        return self._beat_unit

    @beat_unit.setter
    def beat_unit(self, val):
        if not isinstance(val, QuarterDuration):
            val = QuarterDuration(val)
        self._beat_unit = val
        self._set_xml_beat_unit()
        self._set_xml_beat_unit_dots()

    @property
    def per_minute(self):
        return self._per_minute

    @per_minute.setter
    def per_minute(self, val):
        self._per_minute = val
        self._set_xml_per_minute()

    @property
    def sound(self):
        return self._sound

    def get_xml_beat_dot_objects(self):
        return [ch for ch in self.get_children() if isinstance(ch, XMLBeatUnitDot)]



from unittest import TestCase
from musicscore.chord import Chord
from musicscore.exceptions import AlreadyFinalizedError
from musicscore.midi import Midi
from musicscore.part import Part
from musicscore.quarterduration import QuarterDuration
from musicscore.score import Score


def assert_chord_note_values(chord, expected_values):
    assert [
        (note.midi.value, note.quarter_duration) for note in chord.get_children()
    ] == expected_values


class TestFinalUpdates(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_child(Part("p1"))
        self.measure = self.part.add_measure()
        self.staff = self.measure.add_staff()
        self.voice = self.staff.add_voice()
        self.voice.update_beats()
        self.beats = self.voice.get_children()
        self.chord_1 = self.voice._add_chord(Chord(60, QuarterDuration(1, 3)))[0]
        self.chords_2 = self.voice._add_chord(
            Chord(61, QuarterDuration(2, 3) + QuarterDuration(3))
        )

    @staticmethod
    def check_note_values(chords, indices=None):
        if indices is None:
            indices = range(len(chords))

        for i, chord in zip(indices, chords):
            if i == 0:
                assert_chord_note_values(chord, [(60, 1 / 3)])
            elif i == 1:
                assert_chord_note_values(chord, [(61, 2 / 3)])
            elif i == 2:
                assert_chord_note_values(chord, [(61, 3)])

    def test_chord_finalize(self):
        self.measure._update_divisions()
        assert self.chord_1.get_children() == []
        self.chord_1.add_midi(Midi(70))
        self.chord_1.finalize()
        assert len(self.chord_1.get_children()) == 2
        assert_chord_note_values(self.chord_1, [(60, 1 / 3), (70, 1 / 3)])
        with self.assertRaises(AlreadyFinalizedError):
            self.chord_1.finalize()

        self.chords_2[0].finalize()
        assert_chord_note_values(self.chords_2[0], [(61, 2 / 3)])

    def test_beat_finalize(self):
        self.measure._update_divisions()
        self.beats[0].finalize()
        self.check_note_values(self.beats[0].get_chords(), [0, 1])

        self.beats[1].finalize()
        self.check_note_values(self.beats[1].get_chords(), [2])

        with self.assertRaises(AlreadyFinalizedError):
            self.beats[0].finalize()
        with self.assertRaises(AlreadyFinalizedError):
            self.beats[1].finalize()

    def test_voice_finalize(self):
        self.measure._update_divisions()
        self.voice.finalize()
        self.check_note_values(self.voice.get_chords())
        with self.assertRaises(AlreadyFinalizedError):
            self.voice.finalize()

    def test_staff_finalize(self):
        self.measure._update_divisions()
        self.staff.finalize()
        self.check_note_values(self.staff.get_chords())
        with self.assertRaises(AlreadyFinalizedError):
            self.staff.finalize()
        self.staff.to_string()

    def test_measure_finalize(self):
        self.measure.finalize()
        self.check_note_values(self.voice.get_chords())
        with self.assertRaises(AlreadyFinalizedError):
            self.measure.finalize()

    def test_part_finalize(self):
        self.part.finalize()
        self.check_note_values(self.voice.get_chords())
        with self.assertRaises(AlreadyFinalizedError):
            self.part.finalize()

    def test_score_finalize(self):
        self.score.finalize()
        self.check_note_values(self.voice.get_chords())
        with self.assertRaises(AlreadyFinalizedError):
            self.score.finalize()

    def test_to_string_calls_finalize_voice(self):
        self.measure._update_divisions()
        assert self.voice._finalized is False
        self.voice.to_string()
        assert self.voice._finalized is True

    def test_to_string_calls_finalize_staff(self):
        self.measure._update_divisions()
        assert self.staff._finalized is False
        self.staff.to_string()
        assert self.staff._finalized is True

    def test_to_string_calls_finalize_part(self):
        assert self.part._finalized is False
        self.part.to_string()
        assert self.part._finalized is True

    def test_to_string_calls_finalize_score(self):
        assert self.score._finalized is False
        self.score.to_string()
        assert self.score._finalized is True



from unittest import TestCase
from unittest.mock import patch, Mock

from musicscore.accidental import Accidental
from musicscore.beat import Beat
from musicscore.chord import Chord
from musicscore.measure import Measure
from musicscore.midi import Midi
from musicscore.musictree import MusicTree
from musicscore.note import Note
from musicscore.part import Part
from musicscore.score import Score
from musicscore.staff import Staff
from musicscore.voice import Voice


class TestMusicTree(TestCase):
    @patch.object(Chord, "get_voice_number")
    @patch.object(Chord, "get_staff_number")
    def test_add_child_type(self, mock_get_voice_number, mock_get_staff_number):
        mock_get_voice_number.return_value = 1
        mock_get_staff_number.return_value = None
        s = Score()
        p = Part("P1")
        m = Measure(1)
        st = Staff()
        v = Voice()
        b = Beat()
        c = Chord(60, 1)
        mi = c.add_midi(Midi(60))
        n = Note(midi=mi)
        acc = Accidental()
        assert p == s.add_child(p)
        assert m == p.add_child(m)
        assert st == m.add_child(st)
        assert v == st.add_child(v)
        assert b == v.add_child(b)
        assert c == b.add_child(c)[0]
        assert n == c._add_child(n)
        assert mi == n.add_child(mi)
        assert acc == mi.add_child(acc)

        objects = [s, m, st, v, b, n, mi]
        for parent in objects:
            with self.assertRaises(TypeError):
                parent.add_child(s)
        for parent in objects:
            if parent != s:
                with self.assertRaises(TypeError):
                    parent.add_child(p)
        for parent in objects:
            if parent != p:
                with self.assertRaises(TypeError):
                    parent.add_child(m)

        for child in objects:
            with self.assertRaises(NotImplementedError):
                acc.add_child(child)

    def test_check_args_kwargs(self):
        with self.assertRaises(ValueError):
            MusicTree()._check_args_kwargs(
                args=[1, 2, 3], kwargs={"part_number": 2}, class_name="Score"
            )
        kwargs = MusicTree()._check_args_kwargs(
            args=[1, 2, 3], kwargs={}, class_name="Score", get_class_name="Staff"
        )
        assert kwargs == {"part_number": 1, "measure_number": 2, "staff_number": 3}
        kwargs = MusicTree()._check_args_kwargs(
            args=[],
            kwargs={"part_number": 1, "measure_number": 2, "staff_number": 3},
            class_name="Score",
            get_class_name="Staff",
        )
        assert kwargs == {"part_number": 1, "measure_number": 2, "staff_number": 3}
        with self.assertRaises(ValueError):
            MusicTree()._check_args_kwargs(
                args=[],
                kwargs={"part_number": 1, "measure_number": 2, "beat_number": 3},
                class_name="Score",
            )

        kwargs = MusicTree()._check_args_kwargs(
            args=[1, 2, 3], kwargs={}, class_name="Measure", get_class_name="Beat"
        )
        assert kwargs == {"staff_number": 1, "voice_number": 2, "beat_number": 3}
        with self.assertRaises(ValueError):
            MusicTree()._check_args_kwargs(
                args=[1, 2, 3], kwargs={}, class_name="Measure", get_class_name="Voice"
            )

    def test_get_type_errors(self):
        p = Part("p3")
        m = Measure(1)
        s = Staff()
        v = Voice()
        b = Beat()
        c = Chord(60, 1)
        for object_ in [p, m, s, v, b, c]:
            with self.assertRaises(TypeError):
                object_.get_part()

        for object_ in [m, s, v, b, c]:
            with self.assertRaises(TypeError):
                object_.get_measure()

        for object_ in [s, v, b, c]:
            with self.assertRaises(TypeError):
                object_.get_staff()

        for object_ in [v, b, c]:
            with self.assertRaises(TypeError):
                object_.get_voice()

        for object_ in [b, c]:
            with self.assertRaises(TypeError):
                object_.get_beat()

        for object_ in [c]:
            with self.assertRaises(TypeError):
                object_.get_chord()

    def test_score_get_part(self):
        score = Score()
        assert score.get_part(1) is None
        p1 = score.add_child(Part("p1"))
        assert score.get_part(part_number=1) == p1
        assert score.get_part(part_number=2) is None
        p2 = score.add_child(Part("p2"))
        assert score.get_part(part_number=2) == p2
        assert score.get_part(part_number=3) is None

        with self.assertRaises(ValueError):
            score.get_part(1, 2)
        with self.assertRaises(ValueError):
            score.get_part(staff_number=2)

    def test_part_get_measure(self):
        p = Part("p1")
        assert p.get_measure(1) is None
        m1 = p.add_measure()
        assert p.get_measure(1) == m1
        assert p.get_measure(2) is None
        m2 = p.add_measure()
        assert p.get_measure(1) == m1
        assert p.get_measure(2) == m2
        assert p.get_measure(3) is None

        with self.assertRaises(ValueError):
            p.get_measure(1, 2)

        with self.assertRaises(ValueError):
            p.get_measure(staff_number=2)

    def test_measure_get_staff(self):
        m = Measure(1)
        assert m.get_staff(1) is None
        st1 = m.add_staff()
        assert m.get_staff(1) == st1
        assert m.get_staff(2) is None
        st2 = m.add_staff()
        assert m.get_staff(1) == st1
        assert m.get_staff(2) == st2
        assert m.get_staff(3) is None

        with self.assertRaises(ValueError):
            m.get_staff(1, 2)

        with self.assertRaises(ValueError):
            m.get_staff(voice_number=2)

    @patch("musicscore.measure.Measure")
    def test_staff_get_voice(self, mock_measure):
        st = Staff()
        st._parent = mock_measure
        assert st.get_voice(1) is None
        v1 = st.add_voice()
        assert st.get_voice(1) == v1
        assert st.get_voice(2) is None
        v2 = st.add_voice()
        assert st.get_voice(1) == v1
        assert st.get_voice(2) == v2
        assert st.get_voice(3) is None

        with self.assertRaises(ValueError):
            st.get_voice(1, 2)

        with self.assertRaises(ValueError):
            st.get_voice(staff_number=2)

    def test_voice_get_beat(self):
        v = Voice()
        v._parent = Mock()
        assert v.get_beat(1) is None
        b1 = v.add_beat()
        assert v.get_beat(1) == b1
        assert v.get_beat(2) is None
        b2 = v.add_beat()
        assert v.get_beat(1) == b1
        assert v.get_beat(2) == b2
        assert v.get_beat(3) is None

        with self.assertRaises(ValueError):
            v.get_beat(1, 2)

        with self.assertRaises(ValueError):
            v.get_beat(voice_number=2)

    def test_beat_get_chord(self):
        b = Beat()
        b.get_possible_subdivisions = Mock(return_value=None)
        b._parent = Mock()
        assert b.get_chord(1) is None
        ch1 = b._add_chord(Chord(midis=60, quarter_duration=0.5))[0]
        assert b.get_chord(1) == ch1
        assert b.get_chord(2) is None
        ch2 = b._add_chord(Chord(midis=60, quarter_duration=0.5))[0]
        assert b.get_chord(1) == ch1
        assert b.get_chord(2) == ch2
        assert b.get_chord(3) is None

        with self.assertRaises(ValueError):
            b.get_chord(1, 2)

        with self.assertRaises(ValueError):
            b.get_chord(beat_number=2)



"""
Various time signatures: 2/2 (alla breve), 4/4 (C), 2/2, 3/2, 2/4, 3/4, 4/4, 5/4, 3/8, 6/8, 12/8
"""

from pathlib import Path

from musicscore import Score, generate_measures, Chord, C
from unittest import TestCase


class TestLily11a(TestCase):
    def test_lily_11a_TimeSignatures(self):
        score = Score("Time signatures")
        p = score.add_part("ts")

        for measure in generate_measures(
            [
                (2, 2),
                (4, 4),
                (2, 2),
                (3, 2),
                (2, 4),
                (3, 4),
                (4, 4),
                (5, 4),
                (3, 8),
                (6, 8),
                (12, 8),
            ]
        ):
            p.add_child(measure)
            p.add_chord(Chord(C(5), measure.quarter_duration))

        p.get_measure(1).time.symbol = "cut"
        p.get_measure(2).time.symbol = "common"

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import inspect
from unittest import TestCase
from unittest.mock import patch

from musicscore import Chord, Part, Time, Score
from musicscore.accidental import Accidental
from musicscore.measure import Measure
from musicscore.midi import Midi
from musicscore.note import Note
from musicscore.tests.util import generate_path
from musicxml.xmlelement.xmlelement import XMLPitch, XMLRest, XMLNotehead


class TestMidi(TestCase):
    def test_midi_init(self):
        """
        Test that midi initiation
        """
        with self.assertRaises(TypeError):
            Midi()
        with self.assertRaises(TypeError):
            Midi("80")
        with self.assertRaises(ValueError):
            Midi(11)
        with self.assertRaises(ValueError):
            Midi(128)

        m = Midi(80)
        assert m.accidental.mode == "standard"
        assert m.name == "Ab5"
        assert m.accidental.get_pitch_parameters() == ("A", -1, 5)
        assert isinstance(m.get_pitch_or_rest(), XMLPitch)
        expected = """<pitch>
  <step>A</step>
  <alter>-1</alter>
  <octave>5</octave>
</pitch>
"""

        assert m.get_pitch_or_rest().to_string() == expected
        m = Midi(60)
        expected = """<pitch>
  <step>C</step>
  <octave>4</octave>
</pitch>
"""

        assert m.get_pitch_or_rest().to_string() == expected

    def test_midi_accidental_modes(self):
        m = Midi(60)
        assert m.accidental.get_pitch_parameters() == ("C", 0, 4)
        m.accidental.mode = "enharmonic"
        assert m.accidental.get_pitch_parameters() == ("C", 0, 4)
        m.accidental.mode = "force-flat"
        assert m.accidental.get_pitch_parameters() == ("D", -2, 4)
        m.accidental.mode = "force-sharp"
        assert m.accidental.get_pitch_parameters() == ("B", 1, 3)

    def test_midi_rest(self):
        r = Midi(0)
        assert isinstance(r.get_pitch_or_rest(), XMLRest)
        assert r.accidental.get_pitch_parameters() is None
        assert r.get_pitch_or_rest().to_string() == "<rest />\n"

    @patch("musicscore.chord.Chord", spec=Chord)
    def test_midi_parent_note(self, mock_chord):
        """
        Test if a midi object which is being contained in a note can access it via its parent_note attribute.
        """
        mock_chord.get_staff_number.return_value = None
        mock_chord.number_of_dots = 0
        m = Midi(70)
        m._set_parent_chord(mock_chord)
        assert m.parent_note is None
        n = Note(midi=m)
        assert m.parent_note == n
        with self.assertRaises(TypeError):
            m.parent_note = Measure()

    def test_change_midi_value_or_accidental_mode(self):
        """
        Test if changing midi value changes its pitch or rest
        """
        m = Midi(70)
        expected = """<pitch>
  <step>B</step>
  <alter>-1</alter>
  <octave>4</octave>
</pitch>
"""
        assert m.get_pitch_or_rest().to_string() == expected
        m.value = 69
        expected = """<pitch>
  <step>A</step>
  <octave>4</octave>
</pitch>
"""
        assert m.get_pitch_or_rest().to_string() == expected
        m.accidental.mode = "force-flat"
        expected = """<pitch>
  <step>B</step>
  <alter>-2</alter>
  <octave>4</octave>
</pitch>
"""
        assert m.get_pitch_or_rest().to_string() == expected

        m.value = 0
        expected = """<rest />
"""
        assert m.get_pitch_or_rest().to_string() == expected
        m.value = 61
        m.accidental.mode = "flat"
        expected = """<pitch>
  <step>D</step>
  <alter>-1</alter>
  <octave>4</octave>
</pitch>
"""
        assert m.get_pitch_or_rest().to_string() == expected

    def test_midi_copy(self):
        m = Midi(61, accidental=Accidental(mode="sharp", show=False))
        m.add_tie("start")
        copied = m.__copy__()
        assert m != copied
        assert m.value == copied.value
        assert id(m.accidental) == id(copied.accidental)
        assert id(m._ties) == id(copied._ties)

        copied = m.__deepcopy__()
        assert m != copied
        assert m.value == copied.value
        assert m.accidental != copied.accidental
        assert m.accidental.mode == copied.accidental.mode
        assert m.accidental.show == copied.accidental.show
        assert id(m._ties) != id(copied._ties)
        assert m._ties == copied._ties

        copied = m._copy_for_split()
        assert m != copied
        assert m.value == copied.value
        assert m.accidental != copied.accidental
        assert m.accidental.mode == copied.accidental.mode
        assert m.accidental.show == copied.accidental.show
        assert id(m._ties) != id(copied._ties)
        assert m._ties != copied._ties

    @patch("musicscore.chord.Chord", spec=Chord)
    def test_midi_up_note(self, mock_chord):
        mock_chord.get_staff_number.return_value = None
        mock_chord.number_of_dots = 0
        m = Midi(70)
        m._set_parent_chord(mock_chord)
        n = Note(midi=m)
        assert m.up == n

    def test_midi_tie(self):
        m1 = Midi(60)
        m2 = Midi(60)
        m1.add_tie("start")
        m2.add_tie("stop")
        m2.add_tie("start")
        assert m1._ties == {"start"}
        assert m2._ties == {"start", "stop"}
        m2.remove_tie("start")
        assert m2._ties == {"stop"}
        m2.remove_tie("start")
        m3 = m2.__deepcopy__()
        assert m3.value == m2.value
        assert m3._ties == m2._ties
        m3.add_tie("start")
        assert m3._ties != m2._ties

    def test_is_tied_to_next_and_previous(self):
        m1 = Midi(60)
        m2 = Midi(60)
        m1.add_tie("start")
        m2.add_tie("stop")
        m2.add_tie("start")
        assert not m1.is_tied_to_previous
        assert m1.is_tied_to_next
        assert m2.is_tied_to_previous
        assert m2.is_tied_to_next

    def test_set_staff_number(self):
        midi = Midi(60)
        assert not midi.get_staff_number()
        midi.set_staff_number(2)
        assert midi.get_staff_number() == 2


class TestMidiNoteHead(TestCase):
    def test_notehead_property(self):
        m = Midi(60)
        m.notehead = "square"
        assert isinstance(m.notehead, XMLNotehead)
        assert m.notehead.value_ == "square"

    def test_notehead_after_finalize(self):
        p = Part("p1")
        ch = Chord(60, 1)
        ch.midis[0].notehead = "square"
        p.add_chord(ch)
        p.finalize()
        assert ch.midis[0].parent_note.xml_notehead.value_ == "square"

    def test_notehead_copy_for_split(self):
        midi = Midi(60)
        midi.notehead = "square"
        copied = midi._copy_for_split()
        assert copied.notehead.value_ == "square"

    def test_midi_notehead_after_split(self):
        s = Score()
        p = s.add_part("p1")
        ch = Chord(60, 3)
        ch.midis[0].notehead = "square"
        p.add_measure(Time(2, 4))
        p.add_chord(ch)
        path = generate_path(inspect.currentframe())
        s.export_xml(path)
        assert p.get_chords()[1].midis[0].notehead.value_ == "square"
        assert p.get_chords()[1].midis[0].parent_note.xml_notehead.value_ == "square"



import inspect
from unittest import TestCase

from musicscore import Score, Time, Chord, C
from musicscore.tests.util import generate_path
from musicscore.util import _chord_is_in_a_repetition


# MyXMLTestSuite test_cautionary


class TestShowAccidentalSigns(TestCase):
    def setUp(self):
        super().setUp()
        self.score = Score()
        self.part = self.score.add_part("part")
        self.measure_1 = self.part.add_measure()
        self.measure_2 = self.part.add_measure()
        self.staff_1_1 = self.measure_1.add_staff()
        self.staff_1_2 = self.measure_1.add_staff()
        self.staff_2_1 = self.measure_2.add_staff()
        self.staff_2_2 = self.measure_2.add_staff()
        self.voice_1_1_1 = self.staff_1_1.add_voice()
        self.voice_1_1_2 = self.staff_1_1.add_voice()
        self.voice_2_1_1 = self.staff_2_1.add_voice()
        self.voice_2_1_2 = self.staff_2_1.add_voice()
        all_names = [
            "score",
            "part",
            "measure_1",
            "measure_2",
            "staff_1_1",
            "staff_1_2",
            "staff_2_1",
            "staff_2_2",
            "voice_1_1_2",
            "voice_2_1_1",
            "voice_2_1_2",
        ]
        self.all = [eval(f"self.{name}", {"self": self}) for name in all_names]

    def reset_show_signs(self):
        for obj in self.all:
            obj.show_accidental_signs = None

    def test_show_accidental_signs_property(self):
        for obj in self.all:
            assert obj.show_accidental_signs == "modern"
        self.score.show_accidental_signs = "traditional"
        for obj in self.all:
            assert obj.show_accidental_signs == "traditional"
        self.part.show_accidental_signs = "modern"
        assert self.score.show_accidental_signs == "traditional"
        for obj in self.all:
            if obj != self.score:
                assert obj.show_accidental_signs == "modern"

        self.reset_show_signs()
        self.score.show_accidental_signs = "traditional"
        for obj in self.all:
            assert obj.show_accidental_signs == "traditional"

        self.measure_1.show_accidental_signs = "modern"

        for obj in self.all:
            if obj in [
                self.score,
                self.part,
                self.measure_2,
                self.staff_2_1,
                self.staff_2_2,
                self.voice_2_1_1,
                self.voice_2_1_2,
            ]:
                assert obj.show_accidental_signs == "traditional"
            else:
                assert obj.show_accidental_signs == "modern"

    def test_show_accidental_signs_repetition_signs_simple(self):
        s = Score()
        p = s.add_part("p1")
        p.add_measure(Time(3, 4))
        for qd in 3 * [1]:
            p.add_chord(Chord(C(5, "#"), qd))
        s.export_xml(generate_path(inspect.currentframe()))
        for index, chord in enumerate(p.get_chords()):
            if index == 0:
                assert _chord_is_in_a_repetition(chord) is False
            else:
                assert _chord_is_in_a_repetition(chord) is True

        for index, chord in enumerate(p.get_chords()):
            if index == 0:
                assert chord.midis[0].accidental.show is True
            else:
                assert chord.midis[0].accidental.show is False



from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import (
    XMLF,
    XMLFf,
    XMLFff,
    XMLFfff,
    XMLFffff,
    XMLFfffff,
    XMLFp,
    XMLFz,
    XMLMf,
    XMLMp,
    XMLP,
    XMLPf,
    XMLPp,
    XMLPpp,
    XMLPppp,
    XMLPpppp,
    XMLPppppp,
    XMLRf,
    XMLRfz,
    XMLSf,
    XMLSffz,
    XMLSfp,
    XMLSfpp,
    XMLSfz,
    XMLSfzp,
)

__all__ = ["DYNAMICS", "Dynamics"]

#:
DYNAMICS = {
    "f": XMLF,
    "ff": XMLFf,
    "fff": XMLFff,
    "ffff": XMLFfff,
    "fffff": XMLFffff,
    "ffffff": XMLFfffff,
    "fp": XMLFp,
    "fz": XMLFz,
    "mf": XMLMf,
    "mp": XMLMp,
    "p": XMLP,
    "pf": XMLPf,
    "pp": XMLPp,
    "ppp": XMLPpp,
    "pppp": XMLPppp,
    "ppppp": XMLPpppp,
    "pppppp": XMLPppppp,
    "rf": XMLRf,
    "rfz": XMLRfz,
    "sf": XMLSf,
    "sffz": XMLSffz,
    "sfp": XMLSfp,
    "sfpp": XMLSfpp,
    "sfz": XMLSfz,
    "sfzp": XMLSfzp,
}


class Dynamics(XMLWrapper):
    """
    This is a simple wrapper for dynamics objects. Value must be a key in :obj:`DYNAMICS`.
    """

    def __init__(self, value: str, *args, **kwargs):
        super().__init__()
        self.XMLClass = DYNAMICS[value]
        self._xml_object = self.XMLClass(*args, **kwargs)



"""
Slurs on chorded notes: Only the first note of the chord should get the slur notation. Some applications print out
the slur for all notes – these should be ignored.
"""

from pathlib import Path

from musicscore import Score, Chord, G, C, A, D
from unittest import TestCase
from musicscore.util import slur_chords


class TestLily33g(TestCase):
    def test_lily_33g_Slur_ChordedNotes(self):
        score = Score()
        part = score.add_part("p1")

        chords = [
            Chord(m, 1) for m in [[G(4), C(5), G(5)], [A(4), D(5)], [G(4), D(5)], C(5)]
        ]

        slur_chords(chords[:3], number=1)
        slur_chords(chords[2:], number=2)

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)
